#!/usr/bin/env python3

import os
import sys
import logging

logging.basicConfig(stream=sys.stdout, level=logging.INFO,
        style="{", format="certbot_inwx_hook:{levelname}: {message}")

from .constants import CERTBOT_DOMAIN_ENV, CERTBOT_VALIDATION_ENV
from .inwx_challenge import InwxChallenge

def deploy():
    """Deploys INWX challenge"""
    domain = os.environ.get(CERTBOT_DOMAIN_ENV)
    validation = os.environ.get(CERTBOT_VALIDATION_ENV)
    InwxChallenge(domain, validation).deploy_challenge()

def cleanup():
    """Cleans INWX challenge"""
    domain = os.environ.get(CERTBOT_DOMAIN_ENV)
    validation = os.environ.get(CERTBOT_VALIDATION_ENV)
    InwxChallenge(domain, validation).clean_challenge()



API_URL = "https://api.domrobot.com/xmlrpc/"
DEFAULT_CONFIG_FILE_PATH = "/etc/certbot_inwx_hook.ini"
CONFIG_OVERWRITE_ENVVAR = "CERTBOT_INWX_HOOK_CONFIG_FILE"
CERTBOT_DOMAIN_ENV = "CERTBOT_DOMAIN"
CERTBOT_VALIDATION_ENV = "CERTBOT_VALIDATION"



import logging
import time
import traceback
from configparser import ConfigParser
from os import environ, path
from ast import literal_eval

import dns.resolver
import dns.exception

from .inwx import domrobot
from .constants import (DEFAULT_CONFIG_FILE_PATH, CONFIG_OVERWRITE_ENVVAR, 
    API_URL)

logger = logging.getLogger(__name__)

class InwxChallenge:

    def __init__(self, domain, validation, 
            config_file = DEFAULT_CONFIG_FILE_PATH):
        config_file_locations = [ config_file ]
        if environ.get(CONFIG_OVERWRITE_ENVVAR):
            config_file_locations.insert(0, 
                    environ.get(CONFIG_OVERWRITE_ENVVAR))
        for cfg_file in config_file_locations:
            if path.isfile(cfg_file):
                config = ConfigParser()
                config.read(cfg_file)
                self.debug = config.getboolean("default", "debug",
                        fallback=False)
                self.api = domrobot(API_URL, 
                        debug = self.debug)
                self.api.account.login({
                    "user": config.get("default", "user"),
                    "pass": config.get("default", "password")
                    })
                if "nameservers" in config["default"]:
                    self.nameservers = literal_eval(
                            config.get("default", "nameservers", fallback=[]))
                else:
                    self.nameservers = []
                break
        else:
            logger.error("Please provide a configuration file under "
                    + " or ".join(map(lambda x: "'" + str(x) + "'",
                        config_file_locations)) + ".")
            exit(1)
        self.domain = domain
        self.validation = validation

    def __del__(self):
        if hasattr(self, "recordId") and self.recordId:
            self._clean_challenge()

    def deploy_challenge(self):
        """Creates challenge TXT record"""
        logger.info(f"Creating TXT record for \"{self.domain}\"...")
        tld = ".".join(self.domain.split(".")[-2:])
        name = "_acme-challenge." + self.domain
        response = self.api.nameserver.createRecord({"domain": tld,
            "type": "TXT", "content": self.validation, "name": name})
        recordId = response["resData"]["id"]
        logger.info("TXT record registered...")
        logger.info("Checking if DNS has propagated...")
        for i in range(1, 21):
            if (self._has_dns_propagated() == False and i < 20):
                logger.info(f"Try {i:2}/20 failed: DNS not propagated, "
                    "waiting 30s...")
                time.sleep(30)
            elif (self._has_dns_propagated() == False and i == 20):
                logger.error(f"Try {i:2}/20 failed: DNS not propagated.")
                break
            else:
                break
        logger.info("DNS propagated.")
        return recordId

    def clean_challenge(self):
        """Deletes challenge TXT record"""
        tld = ".".join(self.domain.split(".")[-2:])
        name = "_acme-challenge." + self.domain
        name= ".".join(name.split(".")[:-2])
        response = self.api.nameserver.info({"domain": tld, "name": name})
        record = response["resData"]["record"][0]
        if record["type"] != "TXT":
            logger.error(
                f"Could not find correct record for domain \"{self.domain}\". "
                f"Type mismatch should be \"TXT\", "
                f"but is \"{record['type']}\".")
            return
        if record["content"] != self.validation:
            logger.error(
                f"Could not find correct record for domain \"{self.domain}\". "
                f"Content mismatch should be \"{self.validation}\", "
                f"but is \"{record['content']}\".")
            return
        recordId = record["id"]
        self.api.nameserver.deleteRecord({"id": recordId})
        logger.info(f"Deleted TXT record for \"{self.domain}\".")

    def _has_dns_propagated(self):
        """Checks if the TXT record has propagated."""
        txt_records = []
        name = "_acme-challenge." + self.domain
        try:
            resolver = dns.resolver.Resolver()
            nameservers = self.nameservers
            nameservers.extend(resolver.nameservers)
            resolver.nameservers = nameservers
            dns_response = resolver.query(name, "TXT")
            if hasattr(dns_response.response, "answer"):
                for answer in dns_response.response.answer:
                    for record in answer.items:
                        txt_records.append(record.to_text().replace("\"", ""))
        except dns.exception.DNSException as error:
            if (self.debug):
                traceback.print_exc()
            return False

        for txt_record in txt_records:
            if txt_record == self.validation:
                return True

        return False



import subprocess
from setuptools import setup, find_packages
from codecs import open
from os import path
import re

here = path.abspath(path.dirname(__file__))
readme = path.join(here, 'Readme.md')

# Convert the README to reStructuredText for PyPI if pandoc is available.
# Otherwise, just read it.
try:
    readme = subprocess.check_output(['pandoc', '-f', 'markdown',
        '-t', 'rst', readme]).decode('utf-8')
except:
    with open(readme, encoding='utf-8') as f:
        readme = f.read()


setup(
    name = 'certbot_inwx_hook',
    version = "v1.0",

    license = 'MIT',
    description = "Hook for certbot manual mode and the INWX API",
    long_description = readme,
    url = 'https://github.com/faerbit/certbot_inwx_hook',
    author = "Faerbit",
    author_email = 'faerbit at gmail dot com',
    classifiers = [
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Operating System :: OS Independent',
    ],

    packages = ['certbot_inwx_hook'],
    install_requires = [
        'dnspython',
    ],

    entry_points = {
        'console_scripts': [
            'certbot_inwx_hook_deploy = certbot_inwx_hook.main:deploy',
            'certbot_inwx_hook_cleanup = certbot_inwx_hook.main:cleanup',
        ],
    },

    package_data = {
        "/etc/certbot_inwx_hook.sample.ini": [ "certbot_inwx_hook.sample.ini" ],
    },
)



import xmlrpc.client
from xmlrpc.client import _Method
import urllib.request, urllib.error, urllib.parse
import base64
import struct
import time
import hmac
import hashlib
import platform

class domrobot ():
    def __init__ (self, address, debug = False):
        self.url = address
        self.debug = debug
        self.cookie = None
        self.version = "1.0"

    def __getattr__(self, name):
        return _Method(self.__request, name)

    def __request (self, methodname, params):
        tuple_params = tuple([params[0]])
        requestContent = xmlrpc.client.dumps(tuple_params, methodname)
        if(self.debug == True):
            print("Request: "+str(requestContent))
        headers = { 'User-Agent' : 
                'DomRobot/'+self.version+' Python-v'+platform.python_version(), 
                'Content-Type': 'text/xml',
                'content-length': str(len(requestContent))}
        if(self.cookie!=None):
            headers['Cookie'] = self.cookie

        req = urllib.request.Request(self.url, bytearray(requestContent,
            'ascii'), headers)

        response = urllib.request.urlopen(req)
        responseContent = response.read()

        cookies = response.getheader('Set-Cookie')

        if(self.debug == True):
            print("Response: " + responseContent.decode("utf-8"))
        apiReturn = xmlrpc.client.loads(responseContent)
        apiReturn = apiReturn[0][0]
        if (apiReturn["code"] == 2200 and
                "Authentication error" in apiReturn["msg"]):
            raise NameError("Unable to log in. Check your login configuration")
            return False
        if(apiReturn["code"]!=1000):
            raise NameError(f"There was a problem: {apiReturn['msg']} "
                    f"(Error code {apiReturn['code']})")
            return False

        if(cookies!=None):
                cookies = response.getheader('Set-Cookie')
                self.cookie = cookies
                if(self.debug == True):
                    print(("Cookie: " + self.cookie))
        return apiReturn

